#include <new>
#include <base/math.hpp>

#include <engine/e_server_interface.h>
#include <engine/e_config.h>

#include <game/mapitems.hpp>
#include <game/server/gamecontext.hpp>

#include <game/server/gamemodes/exp/exp.hpp>
#include "character.hpp"
#include "laser.hpp"
#include "projectile.hpp"
#include "door.hpp"



struct INPUT_COUNT
{
	int presses;
	int releases;
};

static INPUT_COUNT count_input(int prev, int cur)
{
	INPUT_COUNT c = {0,0};
	prev &= INPUT_STATE_MASK;
	cur &= INPUT_STATE_MASK;
	int i = prev;
	while(i != cur)
	{
		i = (i+1)&INPUT_STATE_MASK;
		if(i&1)
			c.presses++;
		else
			c.releases++;
	}

	return c;
}


MACRO_ALLOC_POOL_ID_IMPL(CHARACTER, MAX_CLIENTS)

// player
CHARACTER::CHARACTER()
: ENTITY(NETOBJTYPE_CHARACTER)
{
	proximity_radius = phys_size;
}

void CHARACTER::reset()
{
	destroy();
}

bool CHARACTER::spawn(PLAYER *player, vec2 pos, int team)
{
	player_state = PLAYERSTATE_UNKNOWN;
	emote_stop = -1;
	last_action = -1;
	if(game.controller->isEXP())
		active_weapon = WEAPON_HAMMER;
	else
		active_weapon = WEAPON_GUN;
	last_weapon = WEAPON_HAMMER;
	queued_weapon = -1;
	
	this->player = player;
	this->pos = pos;
	this->team = team;
	
	core.reset();
	core.world = &game.world.core;
	core.pos = pos;
	game.world.core.characters[player->client_id] = &core;

	reckoning_tick = 0;
	mem_zero(&sendcore, sizeof(sendcore));
	mem_zero(&reckoningcore, sizeof(reckoningcore));
	
	game.world.insert_entity(this);
	alive = true;
	player->force_balanced = false;

	game.controller->on_character_spawn(this);

	return true;
}

void CHARACTER::destroy()
{
	game.world.core.characters[player->client_id] = 0;
	alive = false;
}

void CHARACTER::set_weapon(int w)
{
	if(w == active_weapon)
		return;
		
	last_weapon = active_weapon;
	queued_weapon = -1;
	active_weapon = w;
	if(active_weapon < 0 || active_weapon >= NUM_WEAPONS)
		active_weapon = 0;
	
	game.create_sound(pos, SOUND_WEAPON_SWITCH);
}

bool CHARACTER::is_grounded()
{
	if(col_check_point((int)(pos.x+phys_size/2), (int)(pos.y+phys_size/2+5)))
		return true;
	if(col_check_point((int)(pos.x-phys_size/2), (int)(pos.y+phys_size/2+5)))
		return true;
	return false;
}


int CHARACTER::handle_ninja()
{
	if(active_weapon != WEAPON_NINJA || frozen)
		return 0;
	
	vec2 direction = normalize(vec2(latest_input.target_x, latest_input.target_y));

	if ((server_tick() - ninja.activationtick) > (data->weapons.ninja.duration * server_tickspeed() / 1000))
	{
		// time's up, return
		weapons[WEAPON_NINJA].got = false;
		active_weapon = last_weapon;
		if(active_weapon == WEAPON_NINJA)
			active_weapon = WEAPON_GUN;
		set_weapon(active_weapon);
		return 0;
	}
	
	// force ninja weapon
	set_weapon(WEAPON_NINJA);

	ninja.currentmovetime--;

	if (ninja.currentmovetime == 0)
	{
		// reset player velocity
		core.vel *= 0.2f;
		//return MODIFIER_RETURNFLAGS_OVERRIDEWEAPON;
	}

	if (ninja.currentmovetime > 0)
	{
		// Set player velocity
		core.vel = ninja.activationdir * data->weapons.ninja.velocity;
		vec2 oldpos = pos;
		move_box(&core.pos, &core.vel, vec2(phys_size, phys_size), 0.0f);
		// reset velocity so the client doesn't predict stuff
		core.vel = vec2(0.0f,0.0f);
		if ((ninja.currentmovetime % 2) == 0)
		{
			//create_smoke(pos);
		}

		// check if we hit anything along the way
		{
			CHARACTER *ents[64];
			vec2 dir = pos - oldpos;
			float radius = phys_size * 2.0f; //length(dir * 0.5f);
			vec2 center = oldpos + dir * 0.5f;
			int num = game.world.find_entities(center, radius, (ENTITY**)ents, 64, NETOBJTYPE_CHARACTER);

			for (int i = 0; i < num; i++)
			{
				// Check if entity is a player
				if (ents[i] == this)
					continue;
				// make sure we haven't hit this object before
				bool balreadyhit = false;
				for (int j = 0; j < numobjectshit; j++)
				{
					if (hitobjects[j] == ents[i])
						balreadyhit = true;
				}
				if (balreadyhit)
					continue;

				// check so we are sufficiently close
				if (distance(ents[i]->pos, pos) > (phys_size * 2.0f))
					continue;

				// hit a player, give him damage and stuffs...
				game.create_sound(ents[i]->pos, SOUND_NINJA_HIT);
				// set his velocity to fast upward (for now)
				if(numobjectshit < 10)
					hitobjects[numobjectshit++] = ents[i];
					
				ents[i]->take_damage(vec2(0,10.0f), data->weapons.ninja.base->damage, player->client_id,WEAPON_NINJA);
			}
		}
		return 0;
	}

	return 0;
}


void CHARACTER::do_weaponswitch()
{
	if(reload_timer != 0) // make sure we have reloaded
		return;
	
	if(queued_weapon == -1) // check for a queued weapon
		return;
	
	if(weapons[WEAPON_NINJA].got) // if we have ninja, no weapon selection is possible
		return;
	
	set_weapon(queued_weapon);
}

void CHARACTER::handle_weaponswitch()
{
	int wanted_weapon = active_weapon;
	if(queued_weapon != -1)
		wanted_weapon = queued_weapon;
	
	// select weapon
	int next = count_input(latest_previnput.next_weapon, latest_input.next_weapon).presses;
	int prev = count_input(latest_previnput.prev_weapon, latest_input.prev_weapon).presses;

	if(next < 128) // make sure we only try sane stuff
	{
		while(next) // next weapon selection
		{
			wanted_weapon = (wanted_weapon+1)%NUM_WEAPONS;
			if(weapons[wanted_weapon].got)
				next--;
		}
	}

	if(prev < 128) // make sure we only try sane stuff
	{
		while(prev) // prev weapon selection
		{
			wanted_weapon = (wanted_weapon-1)<0?NUM_WEAPONS-1:wanted_weapon-1;
			if(weapons[wanted_weapon].got)
				prev--;
		}
	}

	// direct weapon selection
	if(latest_input.wanted_weapon)
		wanted_weapon = input.wanted_weapon-1;
	
	if(wanted_weapon == WEAPON_NINJA)
		wanted_weapon = WEAPON_KAMIKAZE;
	else if(wanted_weapon == WEAPON_RIFLE && weapons[WEAPON_FREEZER].got)
		wanted_weapon = WEAPON_FREEZER;
	
	// check for insane values
	if(wanted_weapon >= 0 && wanted_weapon < NUM_WEAPONS && wanted_weapon != active_weapon && weapons[wanted_weapon].got)
		queued_weapon = wanted_weapon;
	
	do_weaponswitch();
}

void CHARACTER::fire_weapon()
{
	if(frozen)
		return;
	
	if(reload_timer != 0)
		return;
	
	do_weaponswitch();
	
	vec2 direction = normalize(vec2(latest_input.target_x, latest_input.target_y));
	
	bool fullauto = false;
	if(active_weapon == WEAPON_GRENADE || active_weapon == WEAPON_SHOTGUN || active_weapon == WEAPON_RIFLE || active_weapon == WEAPON_FREEZER)
		fullauto = true;
	
	// check if we gonna fire
	bool will_fire = false;
	if(count_input(latest_previnput.fire, latest_input.fire).presses) will_fire = true;
	if(fullauto && (latest_input.fire&1) && weapons[active_weapon].ammo) will_fire = true;
	if(!will_fire)
		return;
	
	// check for ammo
	if(!weapons[active_weapon].ammo)
	{
		// 125ms is a magical limit of how fast a human can click
		reload_timer = 125 * server_tickspeed() / 1000;
		game.create_sound(pos, SOUND_WEAPON_NOAMMO);
		return;
	}
	
	vec2 projectile_startpos = pos+direction*phys_size*0.75f;
	
	switch(active_weapon)
	{
		case WEAPON_HAMMER:
		{
			// reset objects hit
			numobjectshit = 0;
			game.create_sound(pos, SOUND_HAMMER_FIRE);
			
			CHARACTER *ents[64];
			int hits = 0;
			int num = game.world.find_entities(pos+direction*phys_size*0.75f, phys_size*0.5f, (ENTITY**)ents, 64, NETOBJTYPE_CHARACTER);

			for (int i = 0; i < num; i++)
			{
				CHARACTER *target = ents[i];
				if (target == this)
					continue;
					
				// hit a player, give him damage and stuffs...
				vec2 fdir = normalize(ents[i]->pos - pos);

				// set his velocity to fast upward (for now)
				game.create_hammerhit(pos);
				ents[i]->take_damage(vec2(0,-1.0f), data->weapons.hammer.base->damage, player->client_id, active_weapon);
				vec2 dir;
				if (length(target->pos - pos) > 0.0f)
					dir = normalize(target->pos - pos);
				else
					dir = vec2(0,-1);
					
				target->core.vel += normalize(dir + vec2(0,-1.1f)) * 10.0f;
				hits++;
				
				if(ents[i]->frozen && game.controller->is_friendly_fire(player->client_id, ents[i]->player->client_id))
					ents[i]->frozen_timer = (float)server_tick() - 1;
			}
			
			//TURRETS COLLISION
			if(game.controller->isEXP() && !player->is_bot())
			{
				for(int b = 0; b < MAX_TURRETS; b++)
				{
					TURRET * t = &( ((GAMECONTROLLER_EXP *)game.controller)->turrets[b] );
					if(!t->used || t->dead)
						continue;
					
					float len = distance(t->pos, pos+direction*phys_size*0.75f);
					
					if(len < phys_size*0.5f)
					{
						game.create_hammerhit(pos);
						if(data->weapons.hammer.base->damage)
							((GAMECONTROLLER_EXP *)game.controller)->hit_turret(b, player->client_id, data->weapons.hammer.base->damage);
						hits++;
					}
				}
			}
			
			// if we hit anything, we have to wait for the reload
			if(hits)
				reload_timer = server_tickspeed()/3;
			
		} break;

		case WEAPON_GUN:
		{
			PROJECTILE *proj = new PROJECTILE(WEAPON_GUN,
				player->client_id,
				projectile_startpos,
				direction,
				(int)(server_tickspeed()*tuning.gun_lifetime),
				1, 0, 0, -1, WEAPON_GUN);
			
			if(!(game.controller->isEXP() && player->is_bot()))
			{
				// pack the projectile and send it to the client directly
				NETOBJ_PROJECTILE p;
				proj->fill_info(&p);
				msg_pack_start(NETMSGTYPE_SV_EXTRAPROJECTILE, 0);
				msg_pack_int(1);
				for(unsigned i = 0; i < sizeof(NETOBJ_PROJECTILE)/sizeof(int); i++)
					msg_pack_int(((int *)&p)[i]);
				msg_pack_end();
				server_send_msg(player->client_id);
			}
							
			game.create_sound(pos, SOUND_GUN_FIRE);
		} break;
		
		case WEAPON_SHOTGUN:
		{
			int shotspread = 2;

			if(!(game.controller->isEXP() && player->is_bot()))
			{
				msg_pack_start(NETMSGTYPE_SV_EXTRAPROJECTILE, 0);
				msg_pack_int(shotspread*2+1);
			}
			
			for(int i = -shotspread; i <= shotspread; i++)
			{
				float spreading[] = {-0.185f, -0.070f, 0, 0.070f, 0.185f};
				float a = get_angle(direction);
				a += spreading[i+2];
				float v = 1-(abs(i)/(float)shotspread);
				float speed = mix((float)tuning.shotgun_speeddiff, 1.0f, v);
				PROJECTILE *proj = new PROJECTILE(WEAPON_SHOTGUN,
					player->client_id,
					projectile_startpos,
					vec2(cosf(a), sinf(a))*speed,
					(int)(server_tickspeed()*tuning.shotgun_lifetime),
					1, 0, 0, -1, WEAPON_SHOTGUN);
				
				if(!(game.controller->isEXP() && player->is_bot()))
				{
					// pack the projectile and send it to the client directly
					NETOBJ_PROJECTILE p;
					proj->fill_info(&p);
					for(unsigned i = 0; i < sizeof(NETOBJ_PROJECTILE)/sizeof(int); i++)
						msg_pack_int(((int *)&p)[i]);
				}
			}
			
			if(!(game.controller->isEXP() && player->is_bot()))
			{
				msg_pack_end();
				server_send_msg(player->client_id);
			}
			
			game.create_sound(pos, SOUND_SHOTGUN_FIRE);
		} break;

		case WEAPON_GRENADE:
		{
			PROJECTILE *proj = new PROJECTILE(WEAPON_GRENADE,
				player->client_id,
				projectile_startpos,
				direction,
				(int)(server_tickspeed()*tuning.grenade_lifetime),
				1, PROJECTILE::PROJECTILE_FLAGS_EXPLODE, 0, SOUND_GRENADE_EXPLODE, WEAPON_GRENADE);

			if(!(game.controller->isEXP() && player->is_bot()))
			{
				// pack the projectile and send it to the client directly
				NETOBJ_PROJECTILE p;
				proj->fill_info(&p);
				msg_pack_start(NETMSGTYPE_SV_EXTRAPROJECTILE, 0);
				msg_pack_int(1);
				for(unsigned i = 0; i < sizeof(NETOBJ_PROJECTILE)/sizeof(int); i++)
					msg_pack_int(((int *)&p)[i]);
				msg_pack_end();
				server_send_msg(player->client_id);
			}

			game.create_sound(pos, SOUND_GRENADE_FIRE);
		} break;
		
		case WEAPON_RIFLE:
		{
			new LASER(pos, direction, tuning.laser_reach, player->client_id, false, false);
			game.create_sound(pos, SOUND_RIFLE_FIRE);
		} break;
		
		case WEAPON_NINJA:
		{
			attack_tick = server_tick();
			ninja.activationdir = direction;
			ninja.currentmovetime = data->weapons.ninja.movetime * server_tickspeed() / 1000;
			
			//reload_timer = data->weapons.ninja.base->firedelay * server_tickspeed() / 1000 + server_tick();
			
			// reset hit objects
			numobjectshit = 0;
			
			game.create_sound(pos, SOUND_NINJA_FIRE);
		} break;
		
		case WEAPON_KAMIKAZE:
		{
			game.create_explosion(pos, player->client_id, WEAPON_KAMIKAZE, false);
			game.create_sound(pos, SOUND_GRENADE_EXPLODE);
			if(alive)
				die(player->client_id, WEAPON_NINJA);
		} break;
		
		case WEAPON_FREEZER:
		{
			new LASER(pos, direction, tuning.laser_reach, player->client_id, false, true);
			game.create_sound(pos, SOUND_RIFLE_FIRE);
			reload_timer = tuning.freezer_reload*server_tickspeed();
		} break;
	}

	if(weapons[active_weapon].ammo > 0) // -1 == unlimited
		weapons[active_weapon].ammo--;
	attack_tick = server_tick();
	if(!reload_timer)
		reload_timer = data->weapons.id[active_weapon].firedelay * server_tickspeed() / 1000;
}

int CHARACTER::handle_weapons()
{
	vec2 direction = normalize(vec2(latest_input.target_x, latest_input.target_y));

	/*
	if(config.dbg_stress)
	{
		for(int i = 0; i < NUM_WEAPONS; i++)
		{
			weapons[i].got = true;
			weapons[i].ammo = 10;
		}

		if(reload_timer) // twice as fast reload
			reload_timer--;
	} */

	//if(active_weapon == WEAPON_NINJA)
	handle_ninja();


	// check reload timer
	if(reload_timer)
	{
		reload_timer--;
		return 0;
	}
	
	/*
	if (active_weapon == WEAPON_NINJA)
	{
		// don't update other weapons while ninja is active
		return handle_ninja();
	}*/

	// fire weapon, if wanted
	fire_weapon();

	// ammo regen
	int ammoregentime = data->weapons.id[active_weapon].ammoregentime;
	if(ammoregentime)
	{
		// If equipped and not active, regen ammo?
		if (reload_timer <= 0)
		{
			if (weapons[active_weapon].ammoregenstart < 0)
				weapons[active_weapon].ammoregenstart = server_tick();

			if ((server_tick() - weapons[active_weapon].ammoregenstart) >= ammoregentime * server_tickspeed() / 1000)
			{
				// Add some ammo
				if(!game.controller->isEXP() || active_weapon == WEAPON_GUN)
					weapons[active_weapon].ammo = min(weapons[active_weapon].ammo + 1, 10);
				else
					weapons[active_weapon].ammo = min(weapons[active_weapon].ammo + 1, data->weapons.id[active_weapon].maxammo);
				weapons[active_weapon].ammoregenstart = -1;
			}
		}
		else
		{
			weapons[active_weapon].ammoregenstart = -1;
		}
	}
	
	return 0;
}

void CHARACTER::on_predicted_input(NETOBJ_PLAYER_INPUT *new_input)
{
	// check for changes
	if(mem_comp(&input, new_input, sizeof(NETOBJ_PLAYER_INPUT)) != 0)
		last_action = server_tick();
		
	// copy new input
	mem_copy(&input, new_input, sizeof(input));
	num_inputs++;
	
	// or are not allowed to aim in the center
	if(input.target_x == 0 && input.target_y == 0)
		input.target_y = -1;	
}

void CHARACTER::on_direct_input(NETOBJ_PLAYER_INPUT *new_input)
{
	mem_copy(&latest_previnput, &latest_input, sizeof(latest_input));
	mem_copy(&latest_input, new_input, sizeof(latest_input));
	
	if(num_inputs > 2 && team != -1)
	{
		handle_weaponswitch();
		fire_weapon();
	}
	
	mem_copy(&latest_previnput, &latest_input, sizeof(latest_input));
}

void CHARACTER::tick()
{
	if(frozen && (float)server_tick() > frozen_timer)
	{
		frozen = false;
		weapons[WEAPON_NINJA].got = false;
		active_weapon = last_weapon;
		last_weapon = -1;
		queued_weapon = -1;
	}
	
	if(player->is_bot())
		handle_bot();
	
	if(player->force_balanced)
	{
		char buf[128];
		str_format(buf, sizeof(buf), "You were moved to %s due to team balancing", game.controller->get_team_name(team));
		game.send_broadcast(buf, player->client_id);
		
		player->force_balanced = false;
	}
	
	core.input = input;
	core.tick(true, !door_open());
	
	if(frozen)
	{
		if(core.vel.y > 0.0f)
			core.vel = vec2(0.0f, core.vel.y);
		else
			core.vel = vec2(0.0f, 0.0f);
		core.hooked_player = -1;
		core.hook_state = HOOK_RETRACTED;
		core.hook_pos = pos;
	}
	
	float phys_size = 28.0f;
	// handle death-tiles
	if(col_get((int)(pos.x+phys_size/2), (int)(pos.y-phys_size/2))&COLFLAG_DEATH ||
			col_get((int)(pos.x+phys_size/2), (int)(pos.y+phys_size/2))&COLFLAG_DEATH ||
			col_get((int)(pos.x-phys_size/2), (int)(pos.y-phys_size/2))&COLFLAG_DEATH ||
			col_get((int)(pos.x-phys_size/2), (int)(pos.y+phys_size/2))&COLFLAG_DEATH)
	{
		die(player->client_id, WEAPON_WORLD);
	}

	// handle weapons
	handle_weapons();

	player_state = input.player_state;

	// Previnput
	previnput = input;
	return;
}

void CHARACTER::tick_defered()
{
	// advance the dummy
	{
		WORLD_CORE tempworld;
		reckoningcore.world = &tempworld;
		reckoningcore.tick(false, !door_open());
		reckoningcore.move();
		reckoningcore.quantize();
	}
	
	//lastsentcore;
	/*if(!dead)
	{*/
		vec2 start_pos = core.pos;
		vec2 start_vel = core.vel;
		bool stuck_before = test_box(core.pos, vec2(28.0f, 28.0f));
		
		core.move();
		
		int door_col = door_collision();
		if(door_col != -1)
			core.pos = start_pos;
		
		if(door_col == DOOR_TYPE_VERTICAL)
			core.vel.x *= -1.1f;
		else if(door_col == DOOR_TYPE_HORIZONTAL)
			core.vel.y *= -1.1f;
		
		bool stuck_after_move = test_box(core.pos, vec2(28.0f, 28.0f));
		core.quantize();
		bool stuck_after_quant = test_box(core.pos, vec2(28.0f, 28.0f));
		pos = core.pos;
		
		if(!stuck_before && (stuck_after_move || stuck_after_quant))
		{
			dbg_msg("player", "STUCK!!! %d %d %d %f %f %f %f %x %x %x %x", 
				stuck_before,
				stuck_after_move,
				stuck_after_quant,
				start_pos.x, start_pos.y,
				start_vel.x, start_vel.y,
				*((unsigned *)&start_pos.x), *((unsigned *)&start_pos.y),
				*((unsigned *)&start_vel.x), *((unsigned *)&start_vel.y));
		}

		int events = core.triggered_events;
		int mask = cmask_all_except_one(player->client_id);
		
		if(events&COREEVENT_GROUND_JUMP) game.create_sound(pos, SOUND_PLAYER_JUMP, mask);
		
		if(events&COREEVENT_HOOK_ATTACH_PLAYER) game.create_sound(pos, SOUND_HOOK_ATTACH_PLAYER, cmask_all());
		if(events&COREEVENT_HOOK_ATTACH_GROUND) game.create_sound(pos, SOUND_HOOK_ATTACH_GROUND, mask);
		if(events&COREEVENT_HOOK_HIT_NOHOOK) game.create_sound(pos, SOUND_HOOK_NOATTACH, mask);
	//}
	
	if(team == -1)
	{
		pos.x = input.target_x;
		pos.y = input.target_y;
	}
	
	// update the sendcore if needed
	{
		NETOBJ_CHARACTER predicted;
		NETOBJ_CHARACTER current;
		mem_zero(&predicted, sizeof(predicted));
		mem_zero(&current, sizeof(current));
		reckoningcore.write(&predicted);
		core.write(&current);

		// only allow dead reackoning for a top of 3 seconds
		if(reckoning_tick+server_tickspeed()*3 < server_tick() || mem_comp(&predicted, &current, sizeof(NETOBJ_CHARACTER)) != 0)
		{
			reckoning_tick = server_tick();
			sendcore = core;
			reckoningcore = core;
		}
	}
}

bool CHARACTER::increase_health(int amount)
{
	if(health >= 10)
		return false;
	health = clamp(health+amount, 0, 10);
	return true;
}

bool CHARACTER::increase_armor(int amount)
{
	if(armor >= 10)
		return false;
	armor = clamp(armor+amount, 0, 10);
	return true;
}

void CHARACTER::die(int killer, int weapon)
{
	int mode_special = game.controller->on_character_death(this, game.players[killer], weapon);
	
	if(game.controller->isEXP())
	{
		dbg_msg("game", "kill killer=\"%d:%s\" victim=\"%d:%s\" weapon=%d special=%d victim_level=%d",
			killer, (game.players[killer] && game.players[killer]->is_bot() ? "Monster" : server_clientname(killer)),
			player->client_id, (player->is_bot() ? "Monster" : server_clientname(player->client_id)),
			weapon, mode_special, player->bot_level);
	}
	else
	{
		dbg_msg("game", "kill killer='%d:%s' victim='%d:%s' weapon=%d special=%d",
			killer, server_clientname(killer),
			player->client_id, server_clientname(player->client_id), weapon, mode_special);
	}
	
	// send the kill message
	NETMSG_SV_KILLMSG msg;
	msg.killer = killer;
	msg.victim = player->client_id;
	msg.weapon = weapon;
	msg.mode_special = mode_special;
	msg.pack(MSGFLAG_VITAL);
	server_send_msg(-1);
	
	// a nice sound
	game.create_sound(pos, SOUND_PLAYER_DIE);
	
	// this is for auto respawn after 3 secs
	player->die_tick = server_tick();
	
	alive = false;
	game.world.remove_entity(this);
	game.world.core.characters[player->client_id] = 0;
	game.create_death(pos, player->client_id);
	
	// we got to wait 0.5 secs before respawning
	player->respawn_tick = server_tick()+server_tickspeed()/2;
	
	if(game.controller->isEXP() && player->is_bot())
	{
		player->must_remove_bot = true;
		if(game.players[killer] && game.players[killer]->get_character() && !game.players[killer]->is_bot() && !(game.players[killer]->game_exp.weapons & WEAPON_GUN))
		{
			game.players[killer]->game_exp.weapons += WEAPON_GUN;
			game.players[killer]->get_character()->weapons[WEAPON_GUN].got = true;
			game.players[killer]->get_character()->weapons[WEAPON_GUN].ammo = data->weapons.id[WEAPON_GUN].maxammo;
            game.send_weapon_pickup(killer, WEAPON_GUN);
			game.send_chat_target(game.players[killer]->client_id, "Picked up : GUN. Say /items for more info.");
		}
		if(player->bot_level == 4)
		{
			for(int i = 0; i < config.sv_max_clients; i++)
			{
				if(game.players[i] && game.players[i]->game_exp.boss_hitter)
					game.players[i]->game_exp.boss_killer = true;
			}
		}
	}
}

bool CHARACTER::take_damage(vec2 force, int dmg, int from, int weapon)
{
	if(game.controller->isEXP() && player->is_bot() && player->bot_level == 4)
	{
		if(game.players[from])
			game.players[from]->game_exp.boss_hitter = true;
		
		BOSS * boss = &((GAMECONTROLLER_EXP *)game.controller)->boss;
		
		if(boss->shield_active)
		{
			if(weapon == WEAPON_HAMMER)
				boss->shield_health -= 3;
			else
				game.create_sound(pos, SOUND_HOOK_NOATTACH);
			
			if(boss->shield_health <= 12 && boss->shield_icons[2])
			{
				game.world.destroy_entity((ENTITY *)boss->shield_icons[2]);
				boss->shield_icons[2] = NULL;
			}
			if(boss->shield_health <= 6 && boss->shield_icons[1])
			{
				game.world.destroy_entity((ENTITY *)boss->shield_icons[1]);
				boss->shield_icons[1] = NULL;
			}
			if(boss->shield_health <= 0 && boss->shield_icons[0])
			{
				game.world.destroy_entity((ENTITY *)boss->shield_icons[0]);
				boss->shield_icons[0] = NULL;
			}
			
			if(boss->shield_health <= 0)
			{
				boss->shield_active = false;
				boss->shield_timer = (float)server_tick() + 15.0f*server_tickspeed();
				game.create_sound(pos, SOUND_GRENADE_EXPLODE);
				game.create_explosion(pos, -1, WEAPON_WORLD, true);
			}
			
			core.vel += force*0.5f;
			return true;
		}
	}
	
	core.vel += force;
	
	if(game.controller->is_friendly_fire(player->client_id, from) && !config.sv_teamdamage)
		return false;

	// player only inflicts half damage on self
	if(weapon != WEAPON_WORLD && from == player->client_id)
		dmg = max(1, dmg/2);
	
	damage_taken++;

	// create healthmod indicator
	if(server_tick() < damage_taken_tick+25)
	{
		// make sure that the damage indicators doesn't group together
		game.create_damageind(pos, damage_taken*0.25f, dmg);
	}
	else
	{
		damage_taken = 0;
		game.create_damageind(pos, 0, dmg);
	}

	if(dmg)
	{
		if(armor)
		{
			if(dmg > 1)
			{
				health--;
				dmg--;
			}
			
			if(dmg > armor)
			{
				dmg -= armor;
				armor = 0;
			}
			else
			{
				armor -= dmg;
				dmg = 0;
			}
		}
		
		health -= dmg;
	}

	damage_taken_tick = server_tick();

	// do damage hit sound
	if(from >= 0 && from != player->client_id && game.players[from])
		game.create_sound(game.players[from]->view_pos, SOUND_HIT, cmask_one(from));

	// check for death
	if(health <= 0)
	{
		die(from, weapon);
		
		// set attacker's face to happy (taunt!)
		if (from >= 0 && from != player->client_id && game.players[from])
		{
			CHARACTER *chr = game.players[from]->get_character();
			if (chr)
			{
				chr->emote_type = EMOTE_HAPPY;
				chr->emote_stop = server_tick() + server_tickspeed();
			}
		}
		
		return false;
	}
	
	if (dmg > 2)
		game.create_sound(pos, SOUND_PLAYER_PAIN_LONG);
	else
		game.create_sound(pos, SOUND_PLAYER_PAIN_SHORT);
	
	emote_type = EMOTE_PAIN;
	emote_stop = server_tick() + 500 * server_tickspeed() / 1000;
	
	// spawn blood?
	return true;
}

void CHARACTER::snap(int snapping_client)
{
	if(networkclipped(snapping_client))
		return;
	
	NETOBJ_CHARACTER *character = (NETOBJ_CHARACTER *)snap_new_item(NETOBJTYPE_CHARACTER, player->client_id, sizeof(NETOBJ_CHARACTER));
	
	// write down the core
	if(game.world.paused)
	{
		// no dead reckoning when paused because the client doesn't know
		// how far to perform the reckoning
		character->tick = 0;
		core.write(character);
	}
	else
	{
		character->tick = reckoning_tick;
		sendcore.write(character);
	}
	
	// set emote
	if (emote_stop < server_tick())
	{
		emote_type = EMOTE_NORMAL;
		emote_stop = -1;
	}

	character->emote = emote_type;

	character->ammocount = 0;
	character->health = 0;
	character->armor = 0;
	
	if(active_weapon == WEAPON_KAMIKAZE)
		character->weapon = WEAPON_NINJA;
	else if(active_weapon == WEAPON_FREEZER)
		character->weapon = WEAPON_RIFLE;
	else
		character->weapon = active_weapon;
	
	if(player->bot_level == 2 && !frozen)
		attack_tick = server_tick();
	
	character->attacktick = attack_tick;

	character->direction = input.direction;

	if(player->client_id == snapping_client)
	{
		character->health = health;
		character->armor = armor;
		if(weapons[active_weapon].ammo > 0)
			character->ammocount = weapons[active_weapon].ammo;
	}

	if (character->emote == EMOTE_NORMAL)
	{
		if(250 - ((server_tick() - last_action)%(250)) < 5)
			character->emote = EMOTE_BLINK;
	}

	character->player_state = player_state;
}






void CHARACTER::handle_bot()
{
	//Level 1 : Use only hammer and gun, try to get your pos, stupid aim, with attack timer
	//Level 2 : Use only kamikaze
	//Level 3 : Use all weapons, good moves, good aim, without attack timer
	
	bool jumping = (bool)input.jump;
	input.direction = 0;
	input.jump = 0;
	latest_input.fire = 0;
	input.fire = 0;
	
	CHARACTER * target_chr = game.world.closest_character_nobot(pos, 700.0f);
	if(!target_chr) return;
	
	if(player->bot_level == 1)
	{
		//1) weapon :
		if(distance(target_chr->pos, pos) < 64.0f)
			queued_weapon = WEAPON_HAMMER;
		else
			queued_weapon = WEAPON_GUN;
		
		//2) move :
		if(pos.x < target_chr->pos.x)
			input.direction = 1;
		else if(pos.x > target_chr->pos.x)
			input.direction = -1;
		vec2 future_pos;
		future_pos.x = pos.x + input.direction*100;
		future_pos.y = pos.y;
		if((col_intersect_line(pos, future_pos, NULL, NULL, false) || pos.y > target_chr->pos.y) && !jumping)
		{
			if(is_grounded() || core.vel.y > -0.3f)
				input.jump = 1;
		}
		
		//3) aim :
		aim(target_chr->pos);
		
		if(!col_intersect_line(pos, target_chr->pos, NULL, NULL, false))
		{
			//4) fire :
			if(queued_weapon != active_weapon)
				return;
			if(distance(target_chr->pos, pos) < 500.0f
				&& weapons[active_weapon].ammo != 0 && reload_timer == 0
				&& (float)server_tick() - attack_timer > 0)
			{
				latest_input.fire = 1;
				input.fire = 1;
				attack_timer = (float)server_tick() + 0.4f*server_tickspeed();
			}
		}
	}
	
	
	else if(player->bot_level == 2)
	{
		//1) move :
		if(pos.x < target_chr->pos.x)
			input.direction = 1;
		else if(pos.x > target_chr->pos.x)
			input.direction = -1;
		vec2 future_pos;
		future_pos.x = pos.x + input.direction*100;
		future_pos.y = pos.y;
		if((col_intersect_line(pos, future_pos, NULL, NULL, false) || pos.y > target_chr->pos.y) && !jumping)
		{
			if(is_grounded() || core.vel.y > -0.3f)
				input.jump = 1;
		}
		
		//2) aim :
		aim(target_chr->pos);

		//3) fire :
		if(distance(target_chr->pos, pos) < 48.0f)
		{
			latest_input.fire = 1;
			input.fire = 1;
		}
	}
	
	
	else if(player->bot_level == 3)
	{
		//1) weapon :
		bool shotgun = false;
		if(weapons[WEAPON_SHOTGUN].got && weapons[WEAPON_SHOTGUN].ammo != 0)
			shotgun = true;

		float d = distance(pos, target_chr->pos);
		if(d < 64)
			queued_weapon = WEAPON_HAMMER;
		else if(d < 480)
		{
			if(shotgun)
				queued_weapon = WEAPON_SHOTGUN;
			else
				queued_weapon = WEAPON_GUN;
		}
		else
			queued_weapon = WEAPON_GUN;
		
		if(queued_weapon != active_weapon)
			return;
		
		//2) move :
		if(col_intersect_line(pos, target_chr->pos, NULL, NULL, false)) //no enemy visible, walk
		{
			if(core.vel.x > 0)
				input.direction = 1;
			else
				input.direction = -1;
			
			//check for wall
			vec2 future_pos;
			future_pos.x = pos.x + input.direction*100;
			future_pos.y = pos.y;
			if(col_intersect_line(pos, future_pos, NULL, NULL, false))
				input.direction *= -1;
			aim(future_pos);
		}
		else
		{
			vec2 his_pos = target_chr->pos;
			if(active_weapon == WEAPON_SHOTGUN || active_weapon == WEAPON_HAMMER)
			{
				//keep him close
				if(his_pos.x < pos.x)
					input.direction = -1;
				else if(his_pos.x > pos.x)
					input.direction = 1;
				aim(target_chr->pos);
				if(server_tick()%10 == 0)
					input.hook ^= 1;
			}
			
			if(is_grounded() || core.vel.y > -0.3f)
				input.jump = 1;
		
			//3) aim :
			bool fire = false;
			if(active_weapon == WEAPON_HAMMER || active_weapon == WEAPON_SHOTGUN)
			{
				aim(target_chr->pos);
				fire = true;
			}
			else
			{
				vec2 trg_pos = target_chr->pos + target_chr->core.vel;
				float angle = get_angle(vec2(trg_pos.x-pos.x, -(trg_pos.y-pos.y)));
				if(target_chr->pos.x < pos.x)
					angle -= 1.25f*3 * 3.14159265358979/180.0f;
				else
					angle += 1.25f*3 * 3.14159265358979/180.0f;
				aim(pos + vec2(get_dir(angle).x, -get_dir(angle).y)*100);
				fire = true;
			}
			
			//4) fire :
			if(fire && reload_timer == 0 && (weapons[active_weapon].ammo == -1 || weapons[active_weapon].ammo > 0))
			{
				latest_input.fire = 1;
				input.fire = 1;
				attack_timer = 1;
			}
		}
	}
	
	
	else if(player->bot_level == 4)
	{
		BOSS * boss = &((GAMECONTROLLER_EXP *)game.controller)->boss;
		
		float d = distance(pos, target_chr->pos);
		if(d < 64.0f || (float)server_tick() < boss->freezer_timer)
			queued_weapon = WEAPON_HAMMER;
		else
			queued_weapon = WEAPON_FREEZER;
		
		aim(target_chr->pos);
		
		if(queued_weapon != active_weapon)
			return;
		
		if(target_chr->frozen || d < 256.0f)
		{
			if(pos.x < target_chr->pos.x)
				input.direction = 1;
			else if(pos.x > target_chr->pos.x)
				input.direction = -1;
			vec2 future_pos;
			future_pos.x = pos.x + input.direction*100;
			future_pos.y = pos.y;
			if((col_intersect_line(pos, future_pos, NULL, NULL, false) || pos.y > target_chr->pos.y) && !jumping)
			{
				if(is_grounded() || core.vel.y > -0.3f)
					input.jump = 1;
			}
		}
		
		if(!col_intersect_line(pos, target_chr->pos, NULL, NULL, false) && reload_timer == 0 && (weapons[active_weapon].ammo == -1 || weapons[active_weapon].ammo > 0))
		{
			latest_input.fire = 1;
			input.fire = 1;
			if(active_weapon == WEAPON_FREEZER)
				boss->freezer_timer = (float)server_tick() + 5.0f*server_tickspeed();
		}
	}
}

void CHARACTER::aim(vec2 target)
{
	target -= pos;
	latest_input.target_x = target.x;
	latest_input.target_y = target.y;
	input.target_x = target.x;
	input.target_y = target.y;
}



void CHARACTER::freeze()
{
	frozen = true;
	frozen_timer = (float)server_tick() + tuning.freeze_timer*server_tickspeed();
	
	ninja.activationtick = server_tick();
	weapons[WEAPON_NINJA].got = true;
	weapons[WEAPON_NINJA].ammo = -1;
	last_weapon = active_weapon;
	active_weapon = WEAPON_NINJA;
	
	if(player->bot_level == 4)
	{
		BOSS * boss = &((GAMECONTROLLER_EXP *)game.controller)->boss;
		boss->shield_health = 0;
		boss->shield_timer = (float)server_tick() + 15.0f*server_tickspeed();
		game.create_sound(pos, SOUND_GRENADE_EXPLODE);
		game.create_explosion(pos, -1, WEAPON_WORLD, true);
	}
}



bool CHARACTER::door_open()
{
	if(player->is_bot())
		return true;
	
	int i;
	for(i = 0; i < MAX_CLIENTS; i++)
	{
		if(!game.players[i] || !game.players[i]->is_bot() || !game.players[i]->get_character())
			continue;
		if(distance(pos, game.players[i]->get_character()->pos) < 700.0f)
			return false;
	}
	
	return true;
}

int CHARACTER::door_collision()
{
	if(door_open())
		return -1;
	
	for(int d = 0; d < ((GAMECONTROLLER_EXP *)game.controller)->cur_door; d++)
	{
		DOOR * door = &((GAMECONTROLLER_EXP *)game.controller)->doors[d];
		if(!door->laser)
			continue;
		
		vec2 start = door->laser->from;
		vec2 end = door->laser->pos;
		
		if(door->type == DOOR_TYPE_VERTICAL)
		{
			for(int y = start.y; y < end.y; y += 3)
			{
				if(distance(core.pos, vec2(start.x, y)) < 30.0f)
					return DOOR_TYPE_VERTICAL;
			}
			if(distance(core.pos, end) < 30.0f)
				return DOOR_TYPE_VERTICAL;
		}
		else if(door->type == DOOR_TYPE_HORIZONTAL)
		{
			for(int x = start.x; x < end.x; x += 3)
			{
				if(distance(core.pos, vec2(x, start.y)) < 30.0f)
					return DOOR_TYPE_HORIZONTAL;
			}
			if(distance(core.pos, end) < 30.0f)
				return DOOR_TYPE_HORIZONTAL;
		}
	}
	
	return -1;
}



void CHARACTER::teleport(vec2 to)
{
	game.create_playerspawn(core.pos);
	game.create_playerspawn(to);
	core.pos = to;
	
	//release his own hook
	core.hooked_player = -1;
	core.hook_state = HOOK_RETRACTED;
	core.hook_pos = core.pos;
	
	//release the hook of all the players which are hooking the teleported player
	for(int i = 0; i < MAX_CLIENTS; i++)
	{
		if(game.players[i] && game.players[i]->get_character())
		{
			CHARACTER_CORE * c = &game.players[i]->get_character()->core;
			if(c->hooked_player == player->client_id)
			{
				c->hooked_player = -1;
				c->hook_state = HOOK_RETRACTED;
				c->hook_pos = core.pos;
			}
		}
	}
}
