#include <engine/e_server_interface.h> //server_tick, server_tickspeed
#include <engine/e_config.h> //sv_max_clients

#include <game/server/gamecontext.hpp>
#include <game/server/gamemodes/exp/exp.hpp>
#include <game/server/gamemodes/exp/environment.hpp>
#include "pickup.hpp"



PICKUP::PICKUP(int _type, int _subtype)
: ENTITY(NETOBJTYPE_PICKUP)
{
	from_drop = false;
	die_timer = (float)server_tick();
	is_boss_shield = false;
	
	type = _type;
	subtype = _subtype;
	proximity_radius = phys_size;
	
	reset();
	
	game.world.insert_entity(this);
}

void PICKUP::reset()
{
	if(from_drop)
	{
		game.world.destroy_entity(this);
		return;
	}
	
	if(type > POWERUP_NINJA)
		spawntick = -1;
	else if(data->pickups[type].spawndelay > 0)
		spawntick = server_tick() + server_tickspeed() * data->pickups[type].spawndelay;
	else
		spawntick = -1;
}

void PICKUP::tick()
{
	if(is_boss_shield)
		return;
	
	// wait for respawn
	if(spawntick > 0)
	{
		if(server_tick() > spawntick)
		{
			// respawn
			spawntick = -1;
			
			if(type == POWERUP_WEAPON)
				game.create_sound(pos, SOUND_WEAPON_SPAWN);
		}
		else
			return;
	}
	
	if(from_drop)
	{
		if((float)server_tick() > die_timer + tuning.pickup_lifetime*server_tickspeed())
		{
			game.world.destroy_entity(this);
			return;
		}
	}
	
	// Check if a player intersected us
	CHARACTER *chr = game.world.closest_character(pos, 20.0f, 0);
	if(chr && chr->alive)
	{
		// player picked us up, is someone was hooking us, let them go
		int respawntime = -1;
		
		PLAYER * p = chr->player;
		
		int taken = true;
		switch(type)
		{
			case POWERUP_HEALTH:
			{
				if(chr->health < chr->player->max_health())
				{
					game.create_sound(pos, SOUND_PICKUP_HEALTH);
					chr->health += 4;
					if(chr->health > chr->player->max_health()) chr->health = chr->player->max_health();
				}
				else taken = false;
			} break;
			
			case POWERUP_ARMOR:
			{
				if(!p->is_bot() && p->game_exp.armor_max < 10)
				{
					if(p->game_exp.armor_max == 0)
						game.send_chat_target(p->client_id, "Picked up : ARMOR. Say /items for more info.");
					game.send_broadcast("Picked up : ARMOR", p->client_id);
					
					game.create_sound(pos, SOUND_PICKUP_ARMOR);
					p->game_exp.armor_max += 1;
					chr->armor += 1;
				}
				else taken = false;
			} break;
			
			case POWERUP_WEAPON:
			{
				if(!p->is_bot() && subtype >= 0 && subtype < NUM_WEAPONS)
				{
					if(!chr->weapons[subtype].got && !(subtype == WEAPON_RIFLE && (p->game_exp.weapons & WEAPON_FREEZER)))
					{
						char msg[64];
						str_format(msg, sizeof(msg), "Picked up : %s. Say /items for more info.", get_weapon_name(subtype));
						game.send_chat_target(p->client_id, msg);
						char broadcast[64];
						str_format(broadcast, sizeof(broadcast), "Picked up : %s", get_weapon_name(subtype));
						game.send_broadcast(broadcast, p->client_id);
						
						p->get_weapon(subtype);
						
						if(subtype == WEAPON_GRENADE)
							game.create_sound(pos, SOUND_PICKUP_GRENADE);
						else if(subtype == WEAPON_SHOTGUN)
							game.create_sound(pos, SOUND_PICKUP_SHOTGUN);
						else if(subtype == WEAPON_RIFLE || subtype == WEAPON_FREEZER)
							game.create_sound(pos, SOUND_PICKUP_SHOTGUN);
						else if(subtype == WEAPON_KAMIKAZE)
							game.create_sound(pos, SOUND_PICKUP_NINJA);
						
						game.send_weapon_pickup(p->client_id, real_weapon(subtype));
					}
					else taken = false;
				}
				else taken = false;
			} break;
			
			case POWERUP_NINJA:
			{
				// activate ninja on target player
				chr->ninja.activationtick = server_tick();
				chr->weapons[WEAPON_NINJA].got = true;
				chr->weapons[WEAPON_NINJA].ammo = -1;
				chr->last_weapon = chr->active_weapon;
				chr->active_weapon = WEAPON_NINJA;
				game.create_sound(pos, SOUND_PICKUP_NINJA);
				
				// loop through all players, setting their emotes
				ENTITY *ents[64];
				int num = game.world.find_entities(vec2(0, 0), 1000000, ents, 64, NETOBJTYPE_CHARACTER);
				for (int i = 0; i < num; i++)
				{
					CHARACTER *c = (CHARACTER *)ents[i];
					if (c != chr)
					{
						c->emote_type = EMOTE_SURPRISE;
						c->emote_stop = server_tick() + server_tickspeed();
					}
				}
				
				chr->emote_type = EMOTE_ANGRY;
				chr->emote_stop = server_tick() + 1200 * server_tickspeed() / 1000;
			} break;
			
			case POWERUP_LIFE:
			{
				if(!p->is_bot())
				{
					if(p->game_exp.items.lives == 0)
						game.send_chat_target(p->client_id, "Picked up : LIFE. Say /items for more info.");
					
					p->game_exp.items.lives++;
					char broadcast[256];
					str_format(broadcast, sizeof(broadcast), "Picked up : LIFE (%d)", p->game_exp.items.lives);
					game.send_broadcast(broadcast, p->client_id);
					
					game.create_sound(pos, SOUND_PICKUP_HEALTH);
					((GAMECONTROLLER_EXP *)game.controller)->send_items(p->client_id);
				}
				else taken = false;
			} break;
			
			case POWERUP_MINOR_POTION:
			{
				if(!p->is_bot())
				{
					if(p->game_exp.items.minor_potions == 0)
						game.send_chat_target(p->client_id, "Picked up : MINOR POTION. Say /items for more info.");
					
					p->game_exp.items.minor_potions++;
					char broadcast[256];
					str_format(broadcast, sizeof(broadcast), "Picked up : MINOR POTION (%d)", p->game_exp.items.minor_potions);
					game.send_broadcast(broadcast, p->client_id);
					
					game.create_sound(pos, SOUND_PICKUP_HEALTH);
					((GAMECONTROLLER_EXP *)game.controller)->send_items(p->client_id);
				}
				else taken = false;
			} break;
			
			case POWERUP_GREATER_POTION:
			{
				if(!p->is_bot())
				{
					if(p->game_exp.items.greater_potions == 0)
						game.send_chat_target(p->client_id, "Picked up : GREATER POTION. Say /items for more info.");
					
					p->game_exp.items.greater_potions++;
					char broadcast[256];
					str_format(broadcast, sizeof(broadcast), "Picked up : GREATER POTION (%d)", p->game_exp.items.greater_potions);
					game.send_broadcast(broadcast, p->client_id);
					
					game.create_sound(pos, SOUND_PICKUP_HEALTH);
					((GAMECONTROLLER_EXP *)game.controller)->send_items(p->client_id);
				}
				else taken = false;
			} break;
		}
		
		if(game.controller->isEXP())
		{
			if(taken)
			{
				const char * name = (p->is_bot() ? "Monster" : server_clientname(chr->player->client_id));
				dbg_msg("game", "pickup player='%d:%s' item=%d/%d",	chr->player->client_id, name, type, subtype);
				if(from_drop)
					game.world.destroy_entity(this);
				else
				{
					respawntime = data->pickups[type].respawntime;
					spawntick = server_tick() + server_tickspeed() * respawntime;
				}
			}
		}
		else
		{
			if(respawntime >= 0)
			{
				dbg_msg("game", "pickup player='%d:%s' item=%d/%d",
					chr->player->client_id, server_clientname(chr->player->client_id), type, subtype);
				spawntick = server_tick() + server_tickspeed() * respawntime;
			}
		}
	}

	if(type == POWERUP_MINOR_POTION || type == POWERUP_GREATER_POTION)
	{
		if((float)server_tick() > animation_timer)
		{
			int id = -1;
			for(int i=0; i < config.sv_max_clients; i++)
			{
				if(game.players[i])
				{
					id = i;
					break;
				}
			}
			if(id == -1) return;
			game.create_death(pos, id);
			game.create_death(pos, -1);
			float sec = (type == POWERUP_GREATER_POTION ? 0.3 : 0.5);
			animation_timer = (float)server_tick() + sec*server_tickspeed();
		}
	}
}

void PICKUP::snap(int snapping_client)
{
	if(spawntick != -1)
		return;

	NETOBJ_PICKUP *up = (NETOBJ_PICKUP *)snap_new_item(NETOBJTYPE_PICKUP, id, sizeof(NETOBJ_PICKUP));
	up->x = (int)pos.x;
	up->y = (int)pos.y;
	up->type = real_pickup(type);
	up->subtype = real_weapon(subtype);
}





void PICKUP::create_random_from_bot(int lvl)
{
	from_drop = true;
	
	int r = server_tick()%100;
	
	if(lvl == 1)
	{
		if(r < 70)
			type = POWERUP_HEALTH; //70%
		else if(r < 90)
			type = POWERUP_MINOR_POTION; //20%
		else
		{ //10%
			type = POWERUP_WEAPON;
			if(r < 96)
				subtype = WEAPON_SHOTGUN; //6%
			else
				subtype = WEAPON_GRENADE; //4%
		}
	}
	else if(lvl == 2)
	{
		if(r < 35)
			type = POWERUP_HEALTH; //35%
		else if(r < 60)
			type = POWERUP_LIFE; //25%
		else if(r < 80)
			type = POWERUP_MINOR_POTION; //20%
		else if(r < 95)
			type = POWERUP_GREATER_POTION; //15%
		else
		{ //5%
			type = POWERUP_WEAPON;
			subtype = WEAPON_KAMIKAZE;
		}
	}
	else if(lvl == 3)
	{
		if(r < 25)
			type = POWERUP_HEALTH; //25%
		else if(r < 30)
			type = POWERUP_LIFE; //5%
		else if(r < 55)
			type = POWERUP_MINOR_POTION; //25%
		else if(r < 80)
			type = POWERUP_GREATER_POTION; //25%
		else
		{ //20%
			type = POWERUP_WEAPON;
			if(r < 92)
				subtype = WEAPON_SHOTGUN; //12%
			else
				subtype = WEAPON_GRENADE; //8%
		}
	}
	else if(lvl == 4)
	{
		type = POWERUP_WEAPON;
		subtype = WEAPON_FREEZER;
	}
	
	for(int id = 0; id < MAX_CLIENTS; id++)
	{
		if(game.players[id] && !game.players[id]->is_bot())
			snap(id);
	}
}



void PICKUP::create_random_from_turret(int turret_type)
{
	from_drop = true;
	
	int r = server_tick()%100; //r < 75
	
	if(r < 70)
		type = POWERUP_ARMOR; //70%
	else
	{
		type = POWERUP_WEAPON; //5%
		if(turret_type == TURRET_TYPE_LASER)
			subtype = WEAPON_RIFLE;
		else if(turret_type == TURRET_TYPE_GUN)
			subtype = WEAPON_GUN;
	}
	
	for(int id = 0; id < MAX_CLIENTS; id++)
	{
		if(game.players[id] && !game.players[id]->is_bot())
			snap(id);
	}
}



void PICKUP::make_boss_shield()
{
	is_boss_shield = true;
	for(int id = 0; id < MAX_CLIENTS; id++)
	{
		if(game.players[id] && !game.players[id]->is_bot())
			snap(id);
	}
}



const char * PICKUP::get_weapon_name(int wid)
{
	if(wid == WEAPON_HAMMER)
		return "HAMMER";
	else if(wid == WEAPON_GUN)
		return "GUN";
	else if(wid == WEAPON_SHOTGUN)
		return "SHOTGUN";
	else if(wid == WEAPON_GRENADE)
		return "GRENADE";
	else if(wid == WEAPON_RIFLE)
		return "RIFLE";
	else if(wid == WEAPON_NINJA)
		return "NINJA";
	else if(wid == WEAPON_KAMIKAZE)
		return "KAMIKAZE";
	else if(wid == WEAPON_FREEZER)
		return "FREEZER";
	return "?";
}

int PICKUP::real_pickup(int type)
{
	if(type == POWERUP_LIFE)
		type = POWERUP_HEALTH;
	else if(type == POWERUP_MINOR_POTION || type == POWERUP_GREATER_POTION)
		type = POWERUP_HEALTH;
	return type;
}

int PICKUP::real_weapon(int wid)
{
	if(wid == WEAPON_KAMIKAZE)
		wid = WEAPON_NINJA;
	else if(wid == WEAPON_FREEZER)
		wid = WEAPON_RIFLE;
	return wid;
}
